<?PHP
// Copyright (c) 2001-2003 ars Cognita, Inc., all rights reserved
/********************************************************************************

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
   
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
   
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
*******************************************************************************/
/**
 * High-level mindmeld classes
 *
 * This script provides classes and methods for high-level
 * access to mindmeld, such as learning new thoughts, adding
 * symbols, forgetting, etc.
 *
 * @author			$Author: richtl $
 * @version		$Revision: 1378 $
 *
 * @package		mindmeld
 */
global $MM_GLOBALS;

require_once( $MM_GLOBALS['home'] . 'include/utilities.inc' );
require_once( $MM_GLOBALS['includes'] . 'cortex.inc.php' );
require_once( $MM_GLOBALS['includes'] . 'sensory.inc.php' );

/**
* class mind 
*
* The mind class provides methods for high-level access to brain functions, such
* as learning new thoughts, adding symbols to existing thoughts, forgetting, etc.
*
* Applications should interact with the mind class, which manages the underlying
* brain classes.
*
* How to use: To use the mind class, an application needs only to create a 
* mind instance. All methods can be called using parameters to define how they
* should behave.
*/
class mind extends _brainobject {

	/**
	* Class constructor
	*
	* Creates the object, and can optionally accept configuration 
	* and sqlActions objects as arguments. It will load the default 
	* configuration and sqlActions objects if they are not provided.
	*/
	function mind() {
		_brainObject::_brainObject();
	}
	
	/**
	* Add a new thought to the database
	*
	* Creates a new idea (thought, symbols, and neurons) from a thought
	* object and writes it (thought, symbols, and neurons) into the
	* database.
	*
	* We're not doing this in a transaction, since we want to capture
	* any part that is created successfully.
	*
	* @param	string	$summary	Thought summary text
	* @param	string	$detail		Thought detail text or URL.
	* @return	integer	ID of new thought, NULL on error.
	*/
	function learnThought( $summary, $detail, $displayRaw = FALSE ) {
		
		$this->logMsg( ">>> learnThought", MM_ALL );
		$returnCode = NULL;
		
		// Create a thought object and write it to the database
		$thought = new thought();
		$summary = $thought->setSummary( $summary );
		$detail = $thought->setDetail( $detail );
		$displayRaw = $thought->setDisplayRaw( $displayRaw );
		$thoughtId = $thought->setThoughtId( $thought->insertThought() );
		
		if ( !isset ( $thoughtId ) ) {
			$this->logMsg( "Unable to write thought to database", MM_STOP );
			return NULL;
		}
			
		// Thought written successfully
		
		// Parse the summary and detail into symbols
		$symbol = new symbol();
		$symbolList = $symbol->parseSymbols( "$summary $detail", ' ', MM_NO_HTML_SYMBOLS, MM_FREQ_WEIGHTING );
		
		// PLUGIN HOOK: Manipulate parsed symbols list
		$symbolList = $thought->pluginMgr->executePlugins( 'learnThought_addSymbols_prefilter', $symbolList );
		
		$addedSymbols = $symbol->insertSymbol( $symbolList );
		
		// Create the  tsneurons one at a time
		$defDom = $this->readParam( 'default_tsneuron_dominance' );
		$rein = $this->readParam( 'tsneuron_reinforcement' );
		$freqWt = $this->readParam( 'new_tsneuron_freq_weighting' );
		
		foreach ( $symbolList as $rawSymbol => $freq ) {
			
			// Create a neuron object and write it to the database
			$neuron = new tsNeuron();
			$neuron->linkToThought( $thoughtId );
			$neuron->linkToSymbol( $rawSymbol );
			
			// Get the default dominance
			$wtDom = $defDom + ( $freq - 1 ) * ( $freqWt * $rein );
			$this->logMsg( sprintf( "Freq is %s, weight param is %s", $freq, $freqWt ), MM_INFO );
			$this->logMsg( sprintf( "Setting dominance to %s", $wtDom ), MM_INFO );
			$neuron->setDominance( $wtDom );
			
			$addedNeuron = $neuron->insertNeuron();
			$this->logMsg( "Added tsNeuron", MM_INFO );
		}
		
		$this->logMsg( sprintf ( "<<< learnThought (%s)", $thoughtId ), MM_ALL );
		return $thoughtId;
	}
	
	/**
	* Add a new thought to the database
	*
	* Creates a new idea (thought, symbols, and neurons) from a thought
	* object and writes it (thought, symbols, and neurons) into the
	* database.
	*
	* We're not doing this in a transaction, since we want to capture
	* any part that is created successfully.
	*
	* @param	string	$summary	Thought summary text
	* @param	string	$detail		Thought detail text or URL.
	* @param	string	$sumMult		Multiplier for default summary symbol weight
	* @param	string	$detMult		Multiplier for default detail symbol weight
	* @param	string	$miscMult		Multiplier for default misc symbol weight
	* @param	string	$displayRaw	Don't filter data before displaying
	* @return	integer	ID of new thought, NULL on error.
	*/
	function learnThoughtPreweighted( $summary, $detail, $misc = NULL, $sumMult = 1, $detMult = 1, $miscMult = 1, $displayRaw = FALSE ) {
		
		$this->logMsg( ">>> learnThought", MM_ALL );
		$returnCode = NULL;
		
		// Create a thought object and write it to the database
		$thought = new thought();
		$summary = $thought->setSummary( $summary );
		$detail = $thought->setDetail( $detail );
		$displayRaw = $thought->setDisplayRaw( $displayRaw );
		$thoughtId = $thought->setThoughtId( $thought->insertThought() );
		
		if ( !isset ( $thoughtId ) ) {
			$this->logMsg( "Unable to write thought to database" , MM_ERROR );
			return NULL;
		}
			
		// Thought written successfully
		
		// Parse the summary into symbols

		if( !function_exists( 'createNeurons' ) ) {
			/**
			* creates tsNeurons given a list of symbols
			*
			* @param array $symbolList	List of symbols
			* @param integer Symbol neural weight multiplier
			*/
			function createNeurons( $symbolList, $multiplier = 1, $thought ) {
				
				$thoughtId = $thought->getThoughtId();
				
				// Get some default values
        		$defDom = $thought->readParam( 'default_tsneuron_dominance' );
        		$rein = $thought->readParam( 'tsneuron_reinforcement' );
        		$freqWt = $thought->readParam( 'new_tsneuron_freq_weighting' );
        		
        		foreach ( $symbolList as $rawSymbol => $freq ) {
        			
        			// Create a neuron object and write it to the database
        			$neuron = new tsNeuron();
        			$neuron->linkToThought( $thoughtId );
        			$neuron->linkToSymbol( $rawSymbol );
        			
        			// Get the default dominance
        			$wtDom = ( $defDom + ( $freq - 1 ) * ( $freqWt * $rein ) ) * $multiplier ;
        			$thought->logMsg( sprintf( "Freq is %s, weight param is %s", $freq, $freqWt ), MM_INFO );
        			$thought->logMsg( sprintf( "Setting dominance to %s", $wtDom ), MM_INFO );
        			$neuron->setDominance( $wtDom );
        			
        			$addedNeuron = $neuron->insertNeuron();
        			$thought->logMsg( "Added tsNeuron", MM_INFO );
        		}
			}
		}
		
		$symbol = new symbol();
		
		$symbolList = $symbol->parseSymbols( $summary, ' ', MM_NO_HTML_SYMBOLS, MM_FREQ_WEIGHTING );
		// PLUGIN HOOK: Manipulate parsed symbols list
		$symbolList = $thought->pluginMgr->executePlugins( 'learnThought_addSummarySymbols_prefilter', $symbolList );
		$addedSymbols = $symbol->insertSymbol( $symbolList );
		createNeurons( $symbolList, $sumMult, $thought );
		
		$symbolList = $symbol->parseSymbols( $detail, ' ', MM_NO_HTML_SYMBOLS, MM_FREQ_WEIGHTING );
		// PLUGIN HOOK: Manipulate parsed symbols list
		$symbolList = $thought->pluginMgr->executePlugins( 'learnThought_addDetailSymbols_prefilter', $symbolList );
		$addedSymbols = $symbol->insertSymbol( $symbolList );
		createNeurons( $symbolList, $detMult, $thought );
		
		if( isset( $misc ) ) {
    		$symbolList = $symbol->parseSymbols( $misc, ' ', MM_NO_HTML_SYMBOLS, MM_FREQ_WEIGHTING );
    		// PLUGIN HOOK: Manipulate parsed symbols list
			$symbolList = $thought->pluginMgr->executePlugins( 'learnThought_addMiscSymbols_prefilter', $symbolList );
    		$addedSymbols = $symbol->insertSymbol( $symbolList );
    		createNeurons( $symbolList, $detMult, $thought );
		}
			
		$this->logMsg( sprintf ( "<<< learnThought (%s)", $thoughtId ), MM_ALL );
		return $thoughtId;
	}
	
	/**
	* Update an existing thought in the database
	*
	* Creates a new idea (thought, symbols, and neurons) from a thought
	* object and writes it (thought, symbols, and neurons) into the
	* database.
	*
	* We're not doing this in a transaction, since we want to capture
	* any part that is created successfully.
	*
	* @param	integer	$thoughtId	ID of updated thought
	* @param	string	$summary	Thought summary text
	* @param	string	$detail		Thought detail text or URL.
	* @return	integer	ID of new thought, NULL on error.
	*/
	function updateLearnedThought( $thoughtId, $summary, $detail ) {
		
		$this->logMsg( ">>> updateLearnedThought", MM_ALL );
		
		$returnCode = NULL;
		
		// Create a thought object and write it to the database
		$thought = new thought();
		$newId = $thought->updateThought( $thoughtId, $summary, $detail );
		
		if ( !isset ( $newId ) ) {
			$this->logMsg( "Unable to write thought to database", MM_STOP );
			
		} else {
			
			// Thought written successfully
	
			// Parse the summary and detail into symbols
			$symbol = new symbol();
			$symbolList = $symbol->parseSymbols( "$summary $detail", ' ', TRUE, TRUE  );
			$addedSymbols = $symbol->insertSymbol( $symbolList );
			$dbconn = &$symbol->dbconn;
			
			// Create the tsneurons one at a time
			$defDom = $this->readParam( 'default_tsneuron_dominance' );
			$rein = $this->readParam( 'tsneuron_reinforcement' );
			$freqWt = $this->readParam( 'new_tsneuron_freq_weighting' );
			
			foreach ( $symbolList as $rawSymbol => $freq ) {
								
				// Check for existing neuron
				$qSymbol = $dbconn->qstr( $rawSymbol );
				$sql = "SELECT thought, symbol FROM TsNeuron WHERE thought = $thoughtId "
					."AND RTRIM(symbol) = $qSymbol";
				$result = $dbconn->Execute( $sql ); 
				if( $result->RowCount() <> 0 ) {
					$this->logMsg( "tsNeuron already exists", MM_INFO );
				} else {
					// Create a neuron object and write it to the database
					$neuron = new tsNeuron();
					$neuron->linkToThought( $thoughtId );
					$neuron->linkToSymbol( $rawSymbol );
					
					// Get the default dominance
					$wtDom = $defDom + ( $freq - 1 ) * ( $freqWt * $rein );
					$this->logMsg( sprintf( "Freq is %s, weight param is %s", $freq, $freqWt ), MM_INFO );
					$this->logMsg( sprintf( "Setting dominance to %s", $wtDom ), MM_INFO );
					$neuron->setDominance( $wtDom );
					
					$addedNeuron = $neuron->insertNeuron();
					$this->logMsg( "Added tsNeuron", MM_INFO );	
				}
			}
			$returnCode = $thoughtId;
		}
		$this->logMsg( sprintf ( "<<< updateLearnedThought (%s)", $returnCode ), MM_ALL );
		return $returnCode;
	}
	
	/**
	* Add a new symbol and neuron to an existing thought.
	*
	* Adds a new symbol and neuron to an existing thought in the brain.
	* Note that this method is NOT used when a new thought is being added
	* to the brain, only when symbols and neurons need to be added to a
	* thought that already exists.
	*
	* @param	integer	$thoughtId	ID of existing thought
    * @param    string	$symbol     Symbol to be added
	* @return	integer	ID of thought that symbol was added to, else NULL 
	*/
	function learnSymbol( $thoughtId = NULL, $symbol = NULL ) {
		
		$this->logMsg( sprintf ( ">>> learnSymbol (%s,%s)", $thoughtId, $symbol ), MM_ALL );
		
		$returnCode = NULL;
		
		if ( !isset ( $symbol ) || !isset ( $thoughtId ) ) {
			$this->logMsg( "learnSymbol is missing arguments", MM_ERROR );
			
		} else {
			$newThought = new thought();
			$newThought->setThoughtId( $thoughtId );
			if ( $newThought->readThoughtQuick( $thoughtId, FALSE ) <> $thoughtId ) {
				$this->logMsg( sprintf ( "Unable to find thought %s", $thoughtId ), MM_WARN );
				
			} else {
				// Create a symbol object
				$newSymbol = new symbol();
				$newSymbol->setSymbol( $symbol );				
				$insertSymbol = $newSymbol->insertSymbol();
				if ( $insertSymbol === NULL ) {
					$this->logMsg( sprintf ( "Unable to insert symbol '%s'", $symbol ), MM_ERROR );
				} else {
					$this->logMsg( sprintf ( "Created symbol '%s', linking to thought '%s'", $symbol, $thoughtId ), MM_INFO );
					$newNeuron = new tsNeuron();
					$newNeuron->linkToThought( $thoughtId );
					$newNeuron->linkToSymbol( $insertSymbol );
					$insertNeuron = $newNeuron->insertNeuron();
					$returnCode = $insertNeuron;
				}
			}
		}
		
		$this->logMsg( sprintf ( "<<< learnSymbol (%s)", $returnCode ), MM_ALL );
		return $returnCode;	
	}
	
	/**
	* Reinforce affected neurons after successful search
	*
	* Reinforces a memory (a set of neurons) after a decision is validated
	* as being the right one.
	*
	* Reinforcing means:
	*
	* 	1. Increasing the dominance of tsNeurons that link the selected
	*		thought with a symbol that matches one of the stimuli.
	*	2. Creating new symbols and tsNeurons for each stimulus that
	*		doesn't match an existing symbol.

	* @param	integer	$queryId	query ID
	* @param	integer	$thoughtId	thought ID (from decision)
	* @param	string	$reinforcement	Type of reinforcement
	* @return	integer	Query Number of affected neurons, else NULL.
	*/
	function reinforceMemory(	$thoughtId = NULL,
								$queryId = NULL,
								$reinforcement = "UNKNOWN" ) {
		
		$this->logMsg( sprintf ( ">>> reinforceMemory (%s, %s, %s)", $thoughtId, $queryId, $reinforcement ), MM_ALL );
		
		$returnCode = NULL;
		
		switch ( $reinforcement ) {
			
			case "POSITIVE":
				
				// Validate we have a valid queryId and thoughtId
				if ( !isset ( $queryId ) || !isset ( $thoughtId ) ) {
					$this->logMsg( sprintf ( "Missing query (%s) or thought (%s)", $queryId, $thoughtId ), MM_ERROR );
				} else {
					$dbConn = &$this->dbconn;
					
					// Setup and perform query
					$attrib = array( 	'%THOUGHTID%' => $thoughtId,
										'%QUERYID%' => $queryId,
										'%REINFORCEMENT%' => $this->readParam( 'tsneuron_reinforcement' ) );
					$recordSet = $this->getAction( 'reinforce memory', $attrib );
					$numUpdated = $dbConn->Affected_Rows();
					
					// Create any missing symbols.
					
					// Setup and perform query
					$attrib = array( 	'%THOUGHTID%' => $thoughtId,
										'%QUERYID%' => $queryId );
					$recordSet = $this->getAction( 'show missing symbols', $attrib );
					
					// Go through the record set and insert symbols where necessary using
					// the learnSymbol method (above).
					while ( !$recordSet->EOF ) {
					
						$stimulus = $recordSet->fields[0];
						
						$newSymbol = new symbol();
						$stimulus = $newSymbol->setSymbol( $stimulus );
						
						$newSymbol->insertSymbol();
						
						$this->logMsg( sprintf ( "Added symbol '%s'", trim( $stimulus) ), MM_INFO );
						//$this->learnSymbol( $thoughtId, $stimulus );
						$recordSet->Movenext();
					}
									
					// Create any missing neurons.
					
					// Setup and perform query
					
					$attrib = array( 	'%THOUGHTID%' => $thoughtId,
										'%QUERYID%' => $queryId );
					$recordSet = $this->getAction( 'show missing neurons', $attrib );
					// Go through the record set and insert tsneurons where necessary using
					// the learnTsNeuron method (above).
					while ( !$recordSet->EOF ) {
					
						$stimulus = $recordSet->fields[0];
						$newNeuron = new tsNeuron();
						//$newNeuron->linkToThought( $thoughtId );
						//$newNeuron->linkToSymbol( $stimulus );
						$neuronId = $newNeuron->insertNeuron( $thoughtId, $stimulus );
						
						$this->logMsg( sprintf ( "Adding neuron '%s'", trim( $neuronId) ), MM_INFO );
						//$this->learnSymbol( $thoughtId, $stimulus );
						$recordSet->Movenext();
					}
					
					$returnCode = $numUpdated;
				}
				break;
				
			case "NEGATIVE":
				$dbConn = $this->dbconn;
				
				if ( ! $queryId ) {
					
					// No query or thought, so run a generic decision cleanu
					$this->logMsg( "Weakening unused memories", MM_INFO );
						
					// Setup and perform query
					$recordSet = $this->getAction( 'weaken unused memories' );
					
				} else {
					
					if ( $thoughtId ) {
						
						// Have a thought, so weaken a specific memory
						$this->logMsg( sprintf ( "Weakening memory for thought %s and query %s",
							$thoughtId, $queryId ), MM_INFO );
							
						// Setup and perform query
						$attrib = array(	'%THOUGHTID%' => $thoughtId,
											'%QUERYID%' => $queryId );
						$recordSet = $this->getAction( 'weaken memories for a thought', $attrib );
					} else {
						
						// No thought specified, so weaken all unsuccessful decisions associated
						// with the query.
						$this->logMsg( sprintf ( "Weakening useless memories for query %s", $queryId ), MM_INFO );
						
						// Setup and perform query
						$attrib = array( '%QUERYID%' => $queryId );
						$recordSet = $this->getAction( 'weaken memories for a query', $attrib );
					}
				}
				
				//$numUpdated = $dbConn->Affected_Rows();
				$numUpdated = 1;
				$returnCode = $numUpdated;
				break;
				
			default:
				// Don't need to do anything
				$returnCode = NULL;
				break;
		}
		
		$this->logMsg( sprintf ( "<<< reinforceMemory (%s)", $returnCode ), MM_ALL );
		return $returnCode;
	}
	
	/**
	* Submit a query for evaluation
	*
	* Submits a new query for evaluation. Creates stimuli from the query
	* text. Calls the database "think" procedure to evaluate the stimuli
	* and populate the decision table.
	*
	* @param string	$queryText	Text of query (question asked)
	* @param string	$email		Searcher's email address
	* @param string	$status		What type of thoughts to find
	* @param string	$metaquery	Unique session identifier
	* @return	integer		ID of submitted query, else NULL
	*/
	function ask( $queryText = NULL, $email = 'UNKNOWN', $status = 'ACTIVE', $metaquery = '' ) {
		
		$this->logMsg( ">>> ask", MM_ALL );
		
		$returnCode = NULL;
		
		// Verify the query is sufficently defined. Although an email
		// address is nice, we really only require the query text.
		if ( !isset ( $queryText ) ) {
			$this->logMsg( "Attempt to insert improperly-defined query", MM_ERROR );
			
		} else {
				
			// Create a query and write it to the database
			$query = new query();
			$queryText = $query->setQueryText( $queryText );
			$queryId = $query->setQueryId( $query->insertQuery( $queryText, NULL, $_COOKIE['search'] ) );
			
			if ( isset( $queryId ) ) {
				
				// Parse query into stimuli
				$stimulus = new symbol();
				$stimulusList = $stimulus->parseSymbols( $queryText );
				
				// PLUGIN HOOK: Manipulate parsed stimulus list
				$stimulusList = $this->pluginMgr->executePlugins( 'queryStimuli_prefilter', $stimulusList );
				
				// Create each stimulus and write it into the database.
				foreach ( $stimulusList as $rawStimulus ) {
					
					// build stimulus object.
					$newStimulus = new stimulus();
					$stimulusText = $newStimulus->setStimulus( $rawStimulus );
					$stimulusQuery = $newStimulus->linkToQuery( $queryId );
					
					// Write stimulus to database.
					$addedStimulus = $newStimulus->insertStimulus();
					if ( $addedStimulus === NULL ) {
						$this->logMsg( sprintf ( "Error inserting stimulus '%s'", $stimulusText ), MM_ERROR );
						$addedStimulus = $stimulusText;
					} else {	
						$this->logMsg( sprintf ( "Added stimulus '%s' to query %s", $addedStimulus, $queryId ), MM_INFO );
					}
				}
				
				// Connect to the database and execute the query
				$dbConn = $this->dbconn;
				
				// Setup and perform query
				$attrib = array( 	'%QUERYID%' => $queryId,
									'%THOUGHTSTATUS%' => $status );
				
				$record = $this->getAction( 'think', $attrib );
				$returnCode = $queryId;
			}
		}
		$this->logMsg( sprintf ( "<<< ask (%s)", $returnCode ), MM_ALL );
		return $returnCode;
	}
	
	/**
	* Fetch the results list for a query
	*
	* Submits a new query for evaluation. Creates stimuli from the query
	* text. Calls the database "think" procedure to evaluate the stimuli
	* and populate the decision table.
	*
	* @param integer	$queryId	Query ID
	* @param string	$status		What type of thoughts to return
	* @param boolean	$cache	Enable or disable query caching
	* @param boolean $forceUpdate	Force recalculation of strength
	* @return array		Array containing results, else NULL
	*/
	function answers( $queryId = NULL, $status = 'ACTIVE', $cache = TRUE, $forceUpdate = FALSE ) {
		
		$this->logMsg( sprintf( ">>> answers (%s)", $queryId ), MM_ALL );
		
		$answers = NULL;
		$dbConn = $this->dbconn;
		if( !isset( $queryId ) or !isset( $status ) ) {
			$this->logMsg( sprintf( "Either queryId (%s) or status (%s) is unset.", $queryId, $status ), MM_INFO );
			return NULL;
		}
		
		// Setup and perform query
		$attrib = array( 	'%QUERYID%' => $queryId,
							'%STATUS%' => $status );
		$recordSet =  $this->getAction( 'fetch answers', $attrib,  $this->readParam( 'max_results_returned' ) );
		
		// Go through the record set 
		$record = 0;
		while ( !$recordSet->EOF ) { 
			
			$thoughtId = $recordSet->fields[0];
			$summary = trim( $recordSet->fields[1] );
			$strength = $recordSet->fields[2];
			$status = $recordSet->fields[3];
			//$certification = $recordSet->fields[4];
			//$visited = $recordSet->fields[5];
			
			$answers[$record] = array( $thoughtId, $summary, $strength, $status );
			$this->logMsg( sprintf ( "Returning query %s answer %s", $queryId, $thoughtId ), MM_INFO );
			
			$record++;
			$recordSet->MoveNext();
		}
		
		$this->logMsg( "<<< answers", MM_ALL );
		return $answers;
	}
	
	/**
	* Create symbols and neurons from an existing thought detail.
	*
	* @param	string	$summary	Thought summary text
	* @param	string	$detail		Thought detail text or URL.
	* @return	integer	ID of new thought, NULL on error.
	*/
	function relearnThought( $thoughtId, $detail ) {
		print "RELEARNING   ";
		set_time_limit( 360 );
		
		global $ADODB_FETCH_MODE;
		$dbconn = $this->dbconn;
		$thought = new thought();
		
		// Strip tags from the detail
		$detail = preg_replace( "/<[^<>]*>/", ' ', $detail );
		
		// Strip non-symbol characters
		$detail =  trim( eregi_replace ( " {2,}", " ",  eregi_replace ( "[^a-zA-Z0-9\']", " ", $detail ) ) );
		
		// Build the symbol list
		$newSymbols= array_unique( split ( "[^a-zA-Z0-9\']", strtoupper( $detail ) ) );
		
		// Fetch the symbols and the neurons, if they exist, then build or reinforce 
		// as necessary.
		$sql = "SELECT Symbol.symbol FROM Symbol, TsNeuron "
			."WHERE Thought = $thoughtId AND Symbol.symbol = TsNeuron.Symbol "
			."ORDER BY Symbol.symbol";
		print "FETCHING   ";
		
		$ADODB_FETCH_MODE = ADODB_FETCH_NUM;
		$result = $dbconn->Execute( $sql );
		$ADODB_FETCH_MODE = ADODB_FETCH_DEFAULT;
		
		// Create an array of all existing TsNeurons linked to this thought
		$existingTsNeurons = array();
		while( !$result->EOF ) {
			$existingTsNeurons[] = $result->fields[0];
			$result->MoveNext();
		}
		
		// Array of new TsNeurons to be added
		$addTsNeurons = array_diff( $newSymbols, $existingTsNeurons );
		
		// Array of existing TsNeurons to be updated
		$updateTsNeurons = array_intersect( $existingTsNeurons, $newSymbols);
		
		// Create an array of Symbols to be added 
		// Check addTsNeurons against existing symbols to see what symbols need to be added
		$where = "(";
		foreach( $addTsNeurons as $testSymbol ) {
				$where .= ",'$testSymbol'";
		}
		$where .= ")";
		$where = preg_replace( "/\(,/", "(", $where );
		
		$sql = "SELECT symbol FROM Symbol WHERE symbol in $where";
		$result = $dbconn->Execute( $sql );
		
		// Array of Symbols that DON'T need to be added
		$existingSymbols = $result->fields;
		
		// Array of Symbols that NEED to be added
		$addSymbols = array();
		foreach( $addTsNeurons as $symbol ) {
			if( !in_array( $symbol, $existingSymbols ) ) {
				$addSymbols[] = $symbol;
			}
		}
		
		print "<H1>New Symbols</H1><PRE>";
		print_r( $newSymbols );
		print "</PRE><H1>Existing tsNeurons</H1><PRE>";
		print_r( $existingTsNeurons );
		print "</PRE><H1>Adding tsNeurons</H1><PRE>";
		print_r( $addTsNeurons );
		print "</PRE><H1>Updating tsNeurons</H1><PRE>";
		print_r( $updateTsNeurons );
		print "</PRE><H1>Existing Symbols</H1><PRE>";
		print_r( $existingSymbols );
		print "</PRE><H1>Adding Symbols</H1><PRE>";
		print_r( $addSymbols );
		print "</PRE>";
		
		// Add symbols that don't exist
		$symbol = new symbol;
		$symbol->insertSymbol( $addSymbols );
		
		// Add tsNeurons and  symbols that need to be added
		foreach( $addTsNeurons as $tsNeuron ) {
			$neuron = new tsNeuron;
			$neuron->insertNeuron( $thoughtId, $tsNeuron );
		}
		
		// Update existing tsNeurons
		//foreach( $updateTsNeurons as $tsNeuron ) {
		//}
	}
	
	
	function relearnDetail( $thoughtId, $detail ) {
		
		$this->logMsg( ">>> learnDetail", MM_ALL );
		
		set_time_limit( 360 );
		
		$returnCode = NULL;
		
		// Create a thought object and write it to the database
		$thought = new thought();
		
		// Remove any non-alphanumerics except apostrophes
		$symbolList = split ( "[^a-zA-Z0-9\']", eregi_replace ( " {2,}", " ", 
			eregi_replace ( "[^a-zA-Z0-9\']", " ", $detail ) ) );
		
		// Create the symbols and tsneuron, one at a time, and then 
		// write each set into the database.
		$neurons = 0;
		foreach ( $symbolList as $rawSymbol ) {
			
			// Create a symbol object and write it to the database
			$symbol = new symbol();
			$requestedSymbol = $symbol->setSymbol( $rawSymbol );
			
			if ( $requestedSymbol != "" ) {
				$addedSymbol = $symbol->insertSymbol();
				
				if ( ! isset( $addedSymbol ) ) {
					$this->logMsg( sprintf ( "Error inserting symbol '%s' (it may already exist)" ),
						$requestedSymbol, MM_WARN );
					// The following allows neurons to be added even if the 
					// symbol is a duplicate and throws an error.
					$addedSymbol = $requestedSymbol;
				} else {					
					$this->logMsg( sprintf ( "Added symbol '%s'", $addedSymbol ), MM_INFO );
				}
				
				// Create a neuron object and write it to the database
				$neuron = new tsNeuron();
				$neuron->linkToThought( $thoughtId );
				$neuron->linkToSymbol( $requestedSymbol );
				$addedNeuron = $neuron->insertNeuron();
				$this->logMsg( "Added tsNeuron", MM_INFO );
				
				// Count the added neurons
				$neurons++;
			}
		}
		$returnCode = $thoughtId;
		$this->logMsg( sprintf ( "<<< relearnDetail (%s)", $returnCode ), MM_ALL );
		return $returnCode;
	}
}
?>
